/* 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version. A copy of the GNU General Public
 License can be found at COPYING.
 
 Many people helped and are helping developing ROCK Linux. Please
 have a look at http://www.rocklinux.org/ and the Documentation/TEAM
 file for details. 
*/

#include <linux/fb.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>

int keeprunning = 1;
int removedelays = 1;
const char* VERSION = "0.1";

void sighandler_int(int nr)
{
	keeprunning = 0;
}

void sighandler_quit(int nr)
{
	removedelays = ! removedelays;
}

void show_usage() {
	printf("\nrawgrab version %s\n\n", VERSION);
	printf("Usage: rawgrab [options]\n");
	printf("-f <fbdev>	select framebuffer device to record	(default: /dev/fb/0)\n");
	printf("-o <outfile>	select raw data output file		(default: grab.raw)\n");
	printf("-m <metafile>	select metatdat data output file	(default: <outfile>.meta)\n\n");
	printf("-h		show this help\n\n");
}
	
int main(int argc, char ** argv)
{
	struct timespec waits;
	FILE *meta;
	struct fb_var_screeninfo var_sinfo;
	int c, rc, xres, yres, bpp, s=0, frmcount=0, opt, fd, out;
	char *fb, *newfb, *oldfb, *tmptr; 
	char	*fbdev = "/dev/fb/0", 
			*outname = "grab.raw", 
			*metaname = NULL;

	while ((opt = getopt (argc, argv, "hf:o:m:")) != -1)
    switch (opt)
      {
      case 'f':
        fbdev = optarg;
        break;
      case 'o':
        outname = optarg;
        break;
      case 'm':
        metaname = optarg;
        break;
      case 'h':
        show_usage();
		  return 0;
        break;
      case '?':
        if (isprint(optopt))
          fprintf (stderr, "unknown option `-%c'.\n", optopt);
        else
          fprintf (stderr, "unknown option character `\\x%x'.\n", optopt);
			show_usage();
        return 1;
      default:
        abort();
      }

	signal(SIGINT, sighandler_int);
	signal(SIGQUIT, sighandler_quit);

	if(strcmp(fbdev,"-") == 0) fd = 0;	
	else {
		fd = open(fbdev, O_RDONLY);
		if ( fd == -1) {
			perror("could not open framebuffer device");
			return 1;
		}
	}
	ioctl(fd, FBIOGET_VSCREENINFO, &var_sinfo);

	if ( var_sinfo.xres_virtual != var_sinfo.xres || var_sinfo.xoffset ||
	     var_sinfo.yres_virtual != var_sinfo.yres || var_sinfo.yoffset ) {
		perror("please turn off virtual screen size using 'fbset'");
		return 1;
	}

	if ( var_sinfo.bits_per_pixel != 15 && var_sinfo.bits_per_pixel != 16 ) {
		perror("please switch to a 15 or 16 pixel video mode using 'fbset'"); 
		return 1;
	}

	if ( var_sinfo.nonstd ) {
		perror("framebuffer is in non-standard video - mode");
		return 1;
	}

	xres = var_sinfo.xres;
	yres = var_sinfo.yres;
	bpp  = var_sinfo.bits_per_pixel;

	fb = mmap(NULL, xres*yres*2, PROT_READ, MAP_SHARED, fd, 0);
	if ( fb == NULL ) {
		perror("creating mmap on framebuffer dev failed");
		return 1;
	}
	/* using mmap, can close fd here */
	close(fd);

	if(strcmp(outname,"-") == 0) out = 1;
	else {
		out = open(outname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
		if ( out == -1) {
			perror("could not open output file");
			return 1;
		}
	}

	newfb = calloc(1, xres*yres*2);
	oldfb = calloc(1, xres*yres*2);

	/* this is a 1/4th sec. */
	waits.tv_nsec = 250000000;
	waits.tv_sec  = 0;

	while ( keeprunning ) {
		tmptr = oldfb;
		oldfb = newfb;
		newfb = tmptr;

		memcpy(newfb, fb, xres*yres*2);
		if ( memcmp(newfb, oldfb, xres*yres*2) ) s=0;

		if ( s++ < 5 || !removedelays ) {
			for (c=0; c<xres*yres*2; c+=rc) {
				frmcount++;
				rc = write(out, fb+c, xres*yres*2 - c);
				if ( rc <= 0 ) {
					perror("can't write raw video file");
					return 1;
				}
			}
			if ( removedelays ) write(1, ".", 1);
			else write(1, "!", 1);
		}
		nanosleep(&waits, NULL);
	}

	/* write metadata */
	if (metaname == NULL) {
		metaname = (char*)malloc(strlen(outname)+6);
		strcpy(metaname,outname);
		strcat(metaname,".meta");
	}
	meta = fopen(metaname,"w+");
	if(meta == NULL) {
		perror("could not open metatdata file");
		return 1;
	}
	fprintf(meta, "%d %d %d %d", xres, yres, bpp, frmcount);
	fclose(meta);

	close(out);
	return 0;
}


